from __future__ import annotations
from typing import Optional, List, Iterable
from collections import namedtuple
import logging
import threading
import binascii
from blatann.nrf import nrf_types, nrf_events
from blatann import gatt
from blatann.uuid import Uuid
from blatann.waitables.event_waitable import IdBasedEventWaitable, EventWaitable
from blatann.exceptions import InvalidOperationException, InvalidStateException
from blatann.event_type import EventSource, Event
from blatann.event_args import *
from blatann.services.ble_data_types import BleDataStream
from blatann.utils.queued_tasks_manager import QueuedTasksManagerBase


logger = logging.getLogger(__name__)


_security_mapping = {
    gatt.SecurityLevel.NO_ACCESS: nrf_types.BLEGapSecModeType.NO_ACCESS,
    gatt.SecurityLevel.OPEN: nrf_types.BLEGapSecModeType.OPEN,
    gatt.SecurityLevel.JUST_WORKS: nrf_types.BLEGapSecModeType.ENCRYPTION,
    gatt.SecurityLevel.MITM: nrf_types.BLEGapSecModeType.MITM,
}


class GattsCharacteristicProperties(gatt.CharacteristicProperties):
    """
    Properties for Gatt Server characeristics
    """
    def __init__(self, read=True, write=False, notify=False, indicate=False, broadcast=False,
                 write_no_response=False, signed_write=False,
                 security_level=gatt.SecurityLevel.OPEN, max_length=20, variable_length=True):
        super(GattsCharacteristicProperties, self).__init__(read, write, notify, indicate, broadcast,
                                                            write_no_response, signed_write)
        self.security_level = security_level
        self.max_len = max_length
        self.variable_length = variable_length


class GattsCharacteristic(gatt.Characteristic):
    """
    Represents a single characteristic within a service. This class is usually not instantiated directly; it
    is added to a service through GattsService::add_characteristic()
    """
    _QueuedChunk = namedtuple("QueuedChunk", ["offset", "data"])

    def __init__(self, ble_device, peer, uuid, properties, notification_manager, value=b"", prefer_indications=True,
                 string_encoding="utf8"):
        """
        :param ble_device:
        :param peer:
        :param uuid:
        :type properties: gatt.GattsCharacteristicProperties
        :type notification_manager: _NotificationManager
        :param value:
        :param prefer_indications:
        """
        super(GattsCharacteristic, self).__init__(ble_device, peer, uuid, properties, string_encoding)
        self._value = value
        self.prefer_indications = prefer_indications
        self._notification_manager = notification_manager
        # Events
        self._on_write = EventSource("Write Event", logger)
        self._on_read = EventSource("Read Event", logger)
        self._on_sub_change = EventSource("Subscription Change Event", logger)
        self._on_notify_complete = EventSource("Notification Complete Event", logger)
        # Subscribed events
        self.ble_device.ble_driver.event_subscribe(self._on_gatts_write, nrf_events.GattsEvtWrite)
        self.ble_device.ble_driver.event_subscribe(self._on_rw_auth_request, nrf_events.GattsEvtReadWriteAuthorizeRequest)
        # Internal state tracking stuff
        self._write_queued = False
        self._read_in_process = False
        self._queued_write_chunks = []
        self.peer.on_disconnect.register(self._on_disconnect)

    """
    Public Methods
    """

    def set_value(self, value, notify_client=False) -> Optional[EventWaitable[GattsCharacteristic, NotificationCompleteEventArgs]]:
        """
        Sets the value of the characteristic.

        :param value: The value to set to. Must be an iterable type such as a str, bytearray, or list of uint8 values.
                      Length must be less than the characteristic's max length.
                      If a str is given, it will be encoded using the string_encoding property.
        :param notify_client: Flag whether or not to notify the client. If indications and notifications are not set up
                              for the characteristic, will raise an InvalidOperationException
        :raises: InvalidOperationException if value length is too long, or notify client set and characteristic
                 is not notifiable
        :return: If notify_client is true, this method will return the waitable for when the notification is sent to the client
        """
        if isinstance(value, BleDataStream):
            value = value.value
        if isinstance(value, str):
            value = value.encode(self.string_encoding)
        if len(value) > self.max_length:
            raise InvalidOperationException("Attempted to set value of {} with length greater than max "
                                            "(got {}, max {})".format(self.uuid, len(value), self.max_length))
        if notify_client and not self.notifiable:
            raise InvalidOperationException("Cannot notify client. "
                                            "{} not set up for notifications or indications".format(self.uuid))

        v = nrf_types.BLEGattsValue(value)
        self.ble_device.ble_driver.ble_gatts_value_set(self.peer.conn_handle, self.value_handle, v)
        self._value = value

        if notify_client and self.client_subscribed and not self._read_in_process:
            return self.notify(None)

    def notify(self, data) -> EventWaitable[GattsCharacteristic, NotificationCompleteEventArgs]:
        """
        Notifies the client with the data provided without setting the data into the characteristic value.
        If data is not provided (None), will notify with the currently-set value of the characteristic

        :param data: The data to notify the client with
        :return: An EventWaitable that will fire when the notification is successfully sent to the client. The waitable
                 also contains the ID of the sent notification which is used in the on_notify_complete event
        :rtype: NotificationCompleteEventWaitable
        """
        if isinstance(data, BleDataStream):
            value = data.value
        if isinstance(data, str):
            value = data.encode(self.string_encoding)
        if not self.notifiable:
            raise InvalidOperationException("Cannot notify client. "
                                            "{} not set up for notifications or indications".format(self.uuid))
        if not self.client_subscribed:
            raise InvalidStateException("Client is not subscribed, cannot notify client")

        notification_id = self._notification_manager.notify(self, self.value_handle, self._on_notify_complete, data)
        return IdBasedEventWaitable(self._on_notify_complete, notification_id)

    """
    Properties
    """

    @property
    def max_length(self) -> int:
        """
        The max possible the value the characteristic can be set to
        """
        return self._properties.max_len

    @property
    def notifiable(self) -> bool:
        """
        Gets if the characteristic is set up to asynchonously notify clients via notifications or indications
        """
        return self._properties.indicate or self._properties.notify

    @property
    def value(self) -> bytes:
        """
        Gets the current value of the characteristic
        """
        return self._value

    @property
    def client_subscribed(self) -> bool:
        """
        Gets if the client is currently subscribed (notify or indicate) to this characteristic
        """
        return self.peer and self.cccd_state != gatt.SubscriptionState.NOT_SUBSCRIBED

    """
    Events
    """

    @property
    def on_write(self) -> Event[GattsCharacteristic, WriteEventArgs]:
        """
        Event generated whenever a client writes to this characteristic.

        EventArgs type: WriteEventArgs

        :return: an Event which can have handlers registered to and deregistered from
        """
        return self._on_write

    @property
    def on_read(self) -> Event[GattsCharacteristic, None]:
        """
        Event generated whenever a client requests to read from this characteristic. At this point, the application
        may choose to update the value of the characteristic to a new value using set_value.

        A good example of this is a "system time" characteristic which reports the applications system time in seconds.
        Instead of updating this characteristic every second, it can be "lazily" updated only when read from.

        NOTE: if there are multiple handlers subscribed to this and each set the value differently, it may cause
        undefined behavior.

        EventArgs type: None

        :return: an Event which can have handlers registered to and deregistered from
        """
        return self._on_read

    @property
    def on_subscription_change(self) -> Event[GattsCharacteristic, SubscriptionStateChangeEventArgs]:
        """
        Event that is generated whenever a client changes its subscription state of the characteristic
        (notify, indicate, none).

        EventArgs type: SubscriptionStateChangeEventArgs

        :return: an Event which can have handlers registered to and deregistered from
        """
        return self._on_sub_change

    @property
    def on_notify_complete(self) -> Event[GattsCharacteristic, NotificationCompleteEventArgs]:
        """
        Event that is generated when a notification or indication sent to the client is successfully sent
        """
        return self._on_notify_complete

    """
    Event Handling
    """

    def _handle_in_characteristic(self, attribute_handle):
        return attribute_handle in [self.value_handle, self.cccd_handle]

    def _execute_queued_write(self, write_op):
        if not self._write_queued:
            return

        self._write_queued = False
        if write_op == nrf_events.BLEGattsWriteOperation.exec_write_req_cancel:
            logger.info("Cancelling write request, char: {}".format(self.uuid))
        else:
            logger.info("Executing write request, char: {}".format(self.uuid))
            # TODO Assume that it was assembled properly. Error handling should go here
            new_value = bytearray()
            for chunk in self._queued_write_chunks:
                new_value += bytearray(chunk.data)
            logger.debug("New value: 0x{}".format(binascii.hexlify(new_value)))
            self.ble_device.ble_driver.ble_gatts_value_set(self.peer.conn_handle, self.value_handle,
                                                           nrf_types.BLEGattsValue(new_value))
            self._value = new_value
            self._on_write.notify(self, WriteEventArgs(self.value))
        self._queued_write_chunks = []

    def _on_cccd_write(self, event):
        """
        :type event: nrf_events.GattsEvtWrite
        """
        self.cccd_state = gatt.SubscriptionState.from_buffer(bytearray(event.data))
        self._on_sub_change.notify(self, SubscriptionStateChangeEventArgs(self.cccd_state))

    def _on_gatts_write(self, driver, event):
        """
        :type event: nrf_events.GattsEvtWrite
        """
        if event.attribute_handle == self.cccd_handle:
            self._on_cccd_write(event)
            return
        elif event.attribute_handle != self.value_handle:
            return
        self._value = bytearray(event.data)
        self._on_write.notify(self, WriteEventArgs(bytes(self.value)))

    def _on_write_auth_request(self, write_event):
        """
        :type write_event: nrf_events.GattsEvtWrite
        """
        if write_event.write_op in [nrf_events.BLEGattsWriteOperation.exec_write_req_cancel,
                                    nrf_events.BLEGattsWriteOperation.exec_write_req_now]:
            self._execute_queued_write(write_event.write_op)
            # Reply should already be handled in database since this can span multiple characteristics and services
            return

        if not self._handle_in_characteristic(write_event.attribute_handle):
            # Handle is not for this characteristic, do nothing
            return

        # Build out the reply
        params = nrf_types.BLEGattsAuthorizeParams(nrf_types.BLEGattStatusCode.success, True,
                                                   write_event.offset, write_event.data)
        reply = nrf_types.BLEGattsRwAuthorizeReplyParams(write=params)

        # Check that the write length is valid
        if write_event.offset + len(write_event.data) > self._properties.max_len:
            params.gatt_status = nrf_types.BLEGattStatusCode.invalid_att_val_length
            self.ble_device.ble_driver.ble_gatts_rw_authorize_reply(write_event.conn_handle, reply)
        else:
            # Send reply before processing write, in case user sets data in gatts_write handler
            try:
                self.ble_device.ble_driver.ble_gatts_rw_authorize_reply(write_event.conn_handle, reply)
            except Exception as e:
                pass
            if write_event.write_op == nrf_events.BLEGattsWriteOperation.prep_write_req:
                self._write_queued = True
                self._queued_write_chunks.append(self._QueuedChunk(write_event.offset, write_event.data))
            elif write_event.write_op in [nrf_events.BLEGattsWriteOperation.write_req,
                                          nrf_types.BLEGattsWriteOperation.write_cmd]:
                self._on_gatts_write(None, write_event)

        # TODO More logic

    def _on_read_auth_request(self, read_event):
        """
        :type read_event: nrf_events.GattsEvtRead
        """
        if not self._handle_in_characteristic(read_event.attribute_handle):
            # Don't care about handles outside of this characteristic
            return

        params = nrf_types.BLEGattsAuthorizeParams(nrf_types.BLEGattStatusCode.success, False, read_event.offset)
        reply = nrf_types.BLEGattsRwAuthorizeReplyParams(read=params)
        if read_event.offset > len(self.value):
            params.gatt_status = nrf_types.BLEGattStatusCode.invalid_offset
        else:
            self._read_in_process = True
            # If the client is reading from the beginning, notify handlers in case an update needs to be made
            if read_event.offset == 0:
                self._on_read.notify(self)
            self._read_in_process = False

        self.ble_device.ble_driver.ble_gatts_rw_authorize_reply(read_event.conn_handle, reply)

    def _on_rw_auth_request(self, driver, event):
        if not self.peer:
            logger.warning("Got RW request when peer not connected: {}".format(event.conn_handle))
            return
        if event.read:
            self._on_read_auth_request(event.read)
        elif event.write:
            self._on_write_auth_request(event.write)
        else:
            logging.error("auth request was not read or write???")

    def _on_disconnect(self, peer, event_args):
        if self.cccd_handle and self.cccd_state != gatt.SubscriptionState.NOT_SUBSCRIBED:
            self.cccd_state = gatt.SubscriptionState.NOT_SUBSCRIBED
            # TODO: Not working goodly
            # self.ble_device.ble_driver.ble_gatts_value_set(nrf_types.BLE_CONN_HANDLE_INVALID, self.cccd_handle,
            #                                                nrf_types.BLEGattsValue(gatt.SubscriptionState.to_buffer(self.cccd_state)))


class GattsService(gatt.Service):
    """
    Represents a registered GATT service that lives locally on the device
    """
    def __init__(self, ble_device, peer, uuid, service_type, notification_manager,
                 start_handle=gatt.BLE_GATT_HANDLE_INVALID, end_handle=gatt.BLE_GATT_HANDLE_INVALID):
        super(GattsService, self).__init__(ble_device, peer, uuid, service_type, start_handle, end_handle)
        self._notification_manager = notification_manager

    @property
    def characteristics(self) -> List[GattsCharacteristic]:
        """
        Gets the list of characteristics in this service
        """
        return self._characteristics

    def add_characteristic(self, uuid: Uuid, properties: GattsCharacteristicProperties,
                           initial_value=b"", prefer_indications=True, string_encoding="utf8"):
        """
        Adds a new characteristic to the service

        :param: The UUID of the characteristic to add
        :param properties: The characteristic's properties
        :param initial_value: The initial value of the characteristic. May be a string, bytearray, or list of ints
        :type initial_value: str or list or bytearray
        :param prefer_indications: Flag for choosing indication/notification if a characteristic has
                                   both indications and notifications available
        :param string_encoding: The encoding method to use when a string value is provided (utf8, ascii, etc.)
        :return: The characteristic just added to the service
        :rtype: GattsCharacteristic
        """
        if isinstance(initial_value, str):
            initial_value = initial_value.encode(string_encoding)
        c = GattsCharacteristic(self.ble_device, self.peer, uuid, properties, self._notification_manager, initial_value,
                                prefer_indications, string_encoding)
        # Register UUID
        self.ble_device.uuid_manager.register_uuid(uuid)

        # Create property structure
        props = nrf_types.BLEGattCharacteristicProperties(properties.broadcast, properties.read, False,
                                                          properties.write, properties.notify, properties.indicate,
                                                          False)
        # Create cccd metadata if notify/indicate enabled
        if properties.notify or properties.indicate:
            cccd_metadata = nrf_types.BLEGattsAttrMetadata(read_auth=False, write_auth=False)
        else:
            cccd_metadata = None

        char_md = nrf_types.BLEGattsCharMetadata(props, cccd_metadata=cccd_metadata)
        security = _security_mapping[properties.security_level]
        attr_metadata = nrf_types.BLEGattsAttrMetadata(security, security, properties.variable_length,
                                                       read_auth=True, write_auth=True)
        attribute = nrf_types.BLEGattsAttribute(uuid.nrf_uuid, attr_metadata, properties.max_len, initial_value)

        handles = nrf_types.BLEGattsCharHandles()  # Populated in call
        self.ble_device.ble_driver.ble_gatts_characteristic_add(self.start_handle, char_md, attribute, handles)

        c.value_handle = handles.value_handle
        c.cccd_handle = handles.cccd_handle

        if c.cccd_handle != gatt.BLE_GATT_HANDLE_INVALID:
            self.end_handle = c.cccd_handle
        else:
            self.end_handle = c.value_handle

        self.characteristics.append(c)
        return c


class GattsDatabase(gatt.GattDatabase):
    """
    Represents the entire GATT server that lives locally on the device which clients read from and write to
    """
    def __init__(self, ble_device, peer):
        super(GattsDatabase, self).__init__(ble_device, peer)
        self.ble_device.ble_driver.event_subscribe(self._on_rw_auth_request,
                                                   nrf_events.GattsEvtReadWriteAuthorizeRequest)
        self._notification_manager = _NotificationManager(ble_device, peer)

    @property
    def services(self) -> List[GattsService]:
        """
        Gets the list of services registered in the database

        :return: list of services in the database
        """
        return self._services

    def iter_services(self) -> Iterable[GattsService]:
        """
        Iterates through all of the registered services in the database

        :return: Generator of the database's services
        :rtype: collections.Iterable[GattsService]
        """
        for s in self.services:
            yield s

    def add_service(self, uuid: Uuid, service_type=gatt.ServiceType.PRIMARY) -> GattsService:
        """
        Adds a service to the local database

        :param uuid: The UUID for the service
        :param service_type: The type of service (primary or secondary)
        :return: The added and newly created service
        """
        # Register UUID
        self.ble_device.uuid_manager.register_uuid(uuid)
        handle = nrf_types.BleGattHandle()
        # Call code to add service to driver
        self.ble_device.ble_driver.ble_gatts_service_add(service_type.value, uuid.nrf_uuid, handle)
        service = GattsService(self.ble_device, self.peer, uuid, service_type, self._notification_manager,
                               handle.handle)
        service.start_handle = handle.handle
        service.end_handle = handle.handle
        self.services.append(service)
        return service

    def clear_pending_notifications(self):
        """
        Clears all pending notifications that are queued to be sent to the client
        """
        self._notification_manager.clear_all()

    def _on_rw_auth_request(self, driver, event):
        """
        :type event: nrf_events.GattsEvtReadWriteAuthorizeRequest
        """
        if not event.write:
            return
        # execute writes can span multiple services and characteristics. Should only reply at the top-level here
        if event.write.write_op not in [nrf_events.BLEGattsWriteOperation.exec_write_req_now,
                                        nrf_events.BLEGattsWriteOperation.exec_write_req_cancel]:
            return
        params = nrf_types.BLEGattsAuthorizeParams(nrf_types.BLEGattStatusCode.success, False)
        reply = nrf_types.BLEGattsRwAuthorizeReplyParams(write=params)
        self.ble_device.ble_driver.ble_gatts_rw_authorize_reply(event.conn_handle, reply)


class _Notification(object):
    _id_counter = 0
    _lock = threading.Lock()

    def __init__(self, characteristic, handle, on_complete, data):
        with _Notification._lock:
            self.id = _Notification._id_counter
            _Notification._id_counter += 1
        self.char = characteristic
        self.handle = handle
        self.on_complete = on_complete
        self.data = data

    def notify_complete(self, reason):
        self.on_complete.notify(self.char, NotificationCompleteEventArgs(self.id, self.data, reason))

    @property
    def type(self):
        if self.char.cccd_state == gatt.SubscriptionState.INDICATION:
            return nrf_types.BLEGattHVXType.indication
        elif self.char.cccd_state == gatt.SubscriptionState.NOTIFY:
            return nrf_types.BLEGattHVXType.notification
        else:
            raise InvalidStateException("Client not subscribed")


class _NotificationManager(QueuedTasksManagerBase):
    """
    Handles queuing of notifications to the client
    """
    def __init__(self, ble_device, peer):
        super(_NotificationManager, self).__init__()
        self.ble_device = ble_device
        self.peer = peer
        self._cur_notification = None
        self.ble_device.ble_driver.event_subscribe(self._on_notify_complete, nrf_events.GattsEvtNotificationTxComplete,
                                                   nrf_events.GattsEvtHandleValueConfirm)
        self.ble_device.ble_driver.event_subscribe(self._on_disconnect, nrf_events.GapEvtDisconnected)
        self.ble_device.ble_driver.event_subscribe(self._on_timeout, nrf_events.GattsEvtTimeout)

    def notify(self, characteristic, handle, event_on_complete, data=None):
        notification = _Notification(characteristic, handle, event_on_complete, data)
        self._add_task(notification)
        return notification.id

    def clear_all(self):
        self._clear_all(NotificationCompleteEventArgs.Reason.QUEUE_CLEARED)

    def _handle_task(self, notification):
        """
        :type notification: _Notification
        """
        if not notification.char.client_subscribed:
            return True
        hvx_params = nrf_types.BLEGattsHvx(notification.handle, notification.type, notification.data)
        self._cur_notification = notification
        self.ble_device.ble_driver.ble_gatts_hvx(self.peer.conn_handle, hvx_params)

    def _handle_task_failure(self, notification, e):
        """
        :type notification: _Notification
        """
        notification.notify_complete(NotificationCompleteEventArgs.Reason.FAILED)

    def _handle_task_cleared(self, notification, reason):
        """
        :type notification: _Notification
        """
        notification.notify_complete(reason)

    def _on_notify_complete(self, driver, event):
        self._cur_notification.notify_complete(NotificationCompleteEventArgs.Reason.SUCCESS)
        self._task_completed(self._cur_notification)

    def _on_disconnect(self, driver, event):
        self._clear_all(NotificationCompleteEventArgs.Reason.CLIENT_DISCONNECTED)

    def _on_timeout(self, driver, event):
        self._clear_all(NotificationCompleteEventArgs.Reason.TIMED_OUT)



"""
This is an example of a broadcaster role BLE device. It advertises as a non-connectable device
and emits the device's current time as a part of the advertising data.
"""
import time
import threading
from blatann import BleDevice
from blatann.gap.advertising import AdvertisingMode, AdvertisingData, AdvertisingFlags
from blatann.examples import example_utils


logger = example_utils.setup_logger(level="DEBUG")


def _get_time_service_data():
    # Get the current time
    t = time.strftime("%H:%M:%S", time.localtime())
    # Service data is 2 bytes UUID + data, use UUID 0x2143
    return "\x43\x21" + t


def wait_for_user_stop(stop_event):
    # Thread that just waits for the user to press enter, then signals the event
    input("Press enter to exit\n")
    stop_event.set()


def main(serial_port):
    stop_event = threading.Event()
    threading.Thread(target=wait_for_user_stop, args=(stop_event, )).start()
    time.sleep(2)

    ble_device = BleDevice(serial_port)
    ble_device.open()

    interval_ms = 100  # Send out an advertising packet every 500ms
    timeout_sec = 0    # Advertise forever
    mode = AdvertisingMode.non_connectable_undirected  # Set mode to not allow connections

    adv_flags = AdvertisingFlags.BR_EDR_NOT_SUPPORTED | AdvertisingFlags.GENERAL_DISCOVERY_MODE
    adv_data = AdvertisingData(flags=adv_flags, local_name="Time", service_data=_get_time_service_data())

    ble_device.advertiser.set_advertise_data(adv_data)
    ble_device.advertiser.start(interval_ms, timeout_sec, auto_restart=True, advertise_mode=mode)

    while True:
        # Update the advertising data every 1 second
        stop_event.wait(1)
        if stop_event.is_set():  # User stopped execution, break out and stop advertising
            break
        # Update the service data and set it in the BLE device
        adv_data.service_data = _get_time_service_data()
        ble_device.advertiser.set_advertise_data(adv_data)

    ble_device.advertiser.stop()
    ble_device.close()


if __name__ == '__main__':
    main("COM8")



from __future__ import annotations
import logging
from blatann.gap.advertise_data import ScanReport, ScanReportCollection
from blatann.nrf import nrf_events, nrf_types
from blatann.waitables import scan_waitable
from blatann.event_type import Event, EventSource

logger = logging.getLogger(__name__)


class ScanParameters(nrf_types.BLEGapScanParams):
    pass


class Scanner(object):
    def __init__(self, ble_device):
        """
        :type ble_device: blatann.device.BleDevice
        """
        self.ble_device = ble_device
        self._default_scan_params = ScanParameters(200, 150, 10)
        self.scanning = False
        ble_device.ble_driver.event_subscribe(self._on_adv_report, nrf_events.GapEvtAdvReport)
        ble_device.ble_driver.event_subscribe(self._on_timeout_event, nrf_events.GapEvtTimeout)
        self.scan_report = ScanReportCollection()
        self._on_scan_received: EventSource[Scanner, ScanReport] = EventSource("On Scan Received", logger)
        self._on_scan_timeout: EventSource[Scanner, ScanReportCollection] = EventSource("On Scan Timeout")

    @property
    def on_scan_received(self) -> Event[Scanner, ScanReport]:
        """
        Event that is triggered whenever a scan report is received

        :return: An event which emits the Scanner and a ScanReport
        """
        return self._on_scan_received

    @property
    def on_scan_timeout(self) -> Event[Scanner, ScanReportCollection]:
        """
        Event that is triggered when scanning completes/times out
        """
        return self._on_scan_timeout

    def set_default_scan_params(self, interval_ms=200, window_ms=150, timeout_seconds=10):
        """
        Sets the default scan parameters so they do not have to be specified each time a scan is started.
        Reference the Bluetooth specification for valid ranges for parameters.

        :param interval_ms: The interval which to scan for advertising packets, in milliseconds
        :param window_ms: How long within a single scan interval to be actively listening for advertising packets,
                          in milliseconds
        :param timeout_seconds: How long to advertise for, in seconds
        """
        self._default_scan_params.interval_ms = interval_ms
        self._default_scan_params.window_ms = window_ms
        self._default_scan_params.timeout_s = timeout_seconds

    def start_scan(self, scan_parameters=None, clear_scan_reports=True):
        """
        Starts a scan and returns a waitable for when the scan completes

        :param scan_parameters: Optional scan parameters. Uses default if not specified
        :type scan_parameters: ScanParameters
        :param clear_scan_reports: Flag to clear out previous scan reports
        :return: A Waitable which will expire once the scan finishes based on the timeout specified.
                 Waitable returns a ScanReportCollection of the advertising packets found
        :rtype: scan_waitable.ScanFinishedWaitable
        """
        self.stop()
        if clear_scan_reports:
            self.scan_report = ScanReportCollection()
        if not scan_parameters:
            scan_parameters = self._default_scan_params
        self.ble_device.ble_driver.ble_gap_scan_start(scan_parameters)
        self.scanning = True
        return scan_waitable.ScanFinishedWaitable(self.ble_device)

    def stop(self):
        """
        Stops an active scan
        """
        if not self.scanning:
            return
        self.scanning = False

        try:
            self.ble_device.ble_driver.ble_gap_scan_stop()
        except:
            pass

    def _on_adv_report(self, driver, adv_report):
        scan_report = self.scan_report.update(adv_report)
        self._on_scan_received.notify(self.ble_device, scan_report)

    def _on_timeout_event(self, driver, event):
        """
        :type event: nrf_events.GapEvtTimeout
        """
        if event.src == nrf_events.BLEGapTimeoutSrc.scan:
            self.scanning = False
            self._on_scan_timeout.notify(self.ble_device, self.scan_report)



import re
import binascii
from blatann.nrf.nrf_types import BLEUUID as _BLEUUID


class Uuid(object):
    def __init__(self, nrf_uuid=None):
        self.nrf_uuid = nrf_uuid

    def __eq__(self, other):
        return str(self) == str(other)


class Uuid128(Uuid):
    def __init__(self, uuid):
        super(Uuid128, self).__init__()
        if isinstance(uuid, str):
            self.uuid_str = uuid.lower()
            self.uuid = self._validate_uuid_str(uuid)
        elif isinstance(uuid, list):
            self.uuid = uuid
            self.uuid_str = self._validate_uuid_list(uuid)
        else:
            raise ValueError("UUID Must be of string or list type")
        self.nrf_uuid = None

    def _validate_uuid_str(self, uuid):
        r = re.compile(r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$")
        if not r.match(uuid):
            raise ValueError("Invalid UUID String. Must be in format of '00112233-aabb-ccdd-eeff-445566778899")
        return binascii.unhexlify(uuid.replace("-", ""))

    def _validate_uuid_list(self, uuid):
        if len(uuid) != 16:
            raise ValueError("UUID Must be 16 bytes long")
        uuid = binascii.hexlify(bytes(uuid)).decode("ascii")
        uuid_sections = uuid[:8], uuid[8:12], uuid[12:16], uuid[16:20], uuid[20:32]

        return "{}-{}-{}-{}-{}".format(*uuid_sections)

    @property
    def uuid_base(self):
        uuid_base = list(self.uuid[:])
        uuid_base[2] = 0
        uuid_base[3] = 0
        return uuid_base

    @property
    def uuid16(self):
        return self.uuid[2] << 8 | self.uuid[3]

    def new_uuid_from_base(self, uuid16):
        if isinstance(uuid16, str):
            uuid16 = int(uuid16, 16)
        elif isinstance(uuid16, Uuid16):
            uuid16 = uuid16.uuid
        if not isinstance(uuid16, int) or uuid16 > 0xFFFF:
            raise ValueError("UUID must be specified as a 16-bit number (0 - 0xFFFF) or a 4 character hex-string")
        uuid = self.uuid_base
        uuid[2] = uuid16 >> 8 & 0xFF
        uuid[3] = uuid16 & 0xFF
        return Uuid128(uuid)

    @classmethod
    def combine_with_base(cls, uuid16, uuid128_base):
        uuid_base = Uuid128(uuid128_base)
        return uuid_base.new_uuid_from_base(uuid16)

    def __str__(self):
        return self.uuid_str


class Uuid16(Uuid):
    def __init__(self, uuid):
        if isinstance(uuid, str):
            uuid = int(uuid, 16)
        if not isinstance(uuid, int) or uuid > 0xFFFF:
            raise ValueError("UUID Must be a valid 16-bit integer")
        super(Uuid16, self).__init__(_BLEUUID(uuid))
        self.uuid = uuid

    def __str__(self):
        return "{:x}".format(self.uuid)





class BondingData(object):
    def __init__(self, bonding_keyset):
        """
        :type bonding_keyset: blatann.nrf.nrf_types.smp.BLEGapSecKeyset
        """
        self.own_ltk = bonding_keyset.own_keys.enc_key
        self.peer_ltk = bonding_keyset.peer_keys.enc_key
        self.peer_id = bonding_keyset.peer_keys.id_key
        self.peer_sign = bonding_keyset.peer_keys.sign_key


class BondDbEntry(object):
    def __init__(self, entry_id=0):
        self.id = entry_id
        self.peer_addr = None
        self.peer_is_client = None
        self.bonding_data = None  # type: BondingData


class BondDatabase(object):
    def create(self):
        """
        :rtype: BondDbEntry
        """
        raise NotImplementedError()

    def add(self, db_entry):
        raise NotImplementedError()

    def update(self, db_entry):
        raise NotImplementedError()

    def delete(self, db_entry):
        raise NotImplementedError()

    def delete_all(self):
        raise NotImplementedError()

    def __iter__(self):
        """
        :rtype: list[BondDbEntry]
        """
        raise NotImplementedError()


class BondDatabaseLoader(object):
    def load(self):
        """
        :rtype: BondDatabase
        """
        raise NotImplementedError()



from blatann.services.current_time.constants import (CURRENT_TIME_SERVICE_UUID,
                                                     CURRENT_TIME_CHARACTERISTIC_UUID,
                                                     LOCAL_TIME_INFO_CHARACTERISTIC_UUID,
                                                     REFERENCE_INFO_CHARACTERISTIC_UUID)
from blatann.services.current_time.data_types import *
from blatann.services.current_time.service import CurrentTimeServer as _CurrentTimeServer


def add_current_time_service(gatts_database, enable_writes=False,
                             enable_local_time_info=False, enable_reference_info=False):
    """
    Adds a Current Time service to the given GATT server database

    :param gatts_database: The database to add the service to
    :type gatts_database: blatann.gatt.gatts.GattsDatabase
    :param enable_writes: Makes the Current time and Local Time info characteristics writable so
                          clients/centrals can update the server's time
    :param enable_local_time_info: Enables the Local Time characteristic in the service
    :param enable_reference_info:  Enables the Reference Info characteristic in the service
    :return: The Current Time service
    :rtype: _CurrentTimeServer
    """
    return _CurrentTimeServer.add_to_database(gatts_database, enable_writes,
                                              enable_local_time_info, enable_reference_info)


class BlatannException(Exception):
    pass


class InvalidStateException(BlatannException):
    pass


class InvalidOperationException(BlatannException):
    pass


class TimeoutError(BlatannException):
    pass


class DecodeError(BlatannException):
    pass



from __future__ import annotations
import binascii
import logging
from blatann.services.battery.constants import *
from blatann.services.battery.data_types import *
from blatann.event_type import EventSource, Event
from blatann.event_args import DecodedReadCompleteEventArgs, ReadCompleteEventArgs, GattOperationCompleteReason
from blatann.waitables import EventWaitable
from blatann.gatt import SecurityLevel, GattStatusCode
from blatann.gatt.gatts import GattsService, GattsCharacteristicProperties


logger = logging.getLogger(__name__)


class BatteryServer(object):
    def __init__(self, service, enable_notifications=False, security_level=SecurityLevel.OPEN):
        """
        :type service: GattsService
        :param enable_notifications:
        :param security_level:
        """
        self._service = service

        battery_level_char_props = GattsCharacteristicProperties(read=True, notify=enable_notifications,
                                                                 variable_length=False,
                                                                 max_length=BatteryLevel.encoded_size(),
                                                                 security_level=security_level)
        self._batt_characteristic = service.add_characteristic(BATTERY_LEVEL_CHARACTERISTIC_UUID,
                                                               battery_level_char_props)

    def set_battery_level(self, battery_percent: int, notify_client=True):
        """
        Sets the new battery level in the service

        :param battery_percent: The new battery percent
        :param notify_client: Whether or not to notify the connected client with the updated value
        """
        battery_percent = int(battery_percent)
        if battery_percent < 0 or battery_percent > 100:
            raise ValueError("Battery % must be between 0 and 100, got {}".format(battery_percent))
        notify_client = notify_client and self._batt_characteristic.notifiable
        return self._batt_characteristic.set_value(BatteryLevel.encode(battery_percent), notify_client)

    @classmethod
    def add_to_database(cls, gatts_database, enable_notifications=False, security_level=SecurityLevel.OPEN):
        """
        :type gatts_database: blatann.gatt.gatts.GattsDatabase
        """
        service = gatts_database.add_service(BATTERY_SERVICE_UUID)
        return BatteryServer(service, enable_notifications, security_level)


class BatteryClient(object):
    def __init__(self, gattc_service):
        """
        :type gattc_service: blatann.gatt.gattc.GattcService
        """
        self._service = gattc_service
        self._batt_characteristic = gattc_service.find_characteristic(BATTERY_LEVEL_CHARACTERISTIC_UUID)
        self._on_battery_level_updated_event = EventSource("Battery Level Update Event")

    def read(self) -> EventWaitable[BatteryClient, DecodedReadCompleteEventArgs[int]]:
        """
        Reads the Battery level characteristic.

        :return: A waitable for when the read completes, which waits for the on_battery_level_update_event to be
                 emitted
        """
        self._batt_characteristic.read().then(self._on_read_complete)
        return EventWaitable(self._on_battery_level_updated_event)

    @property
    def on_battery_level_updated(self) -> Event[BatteryClient, DecodedReadCompleteEventArgs[int]]:
        """
        Event that is generated whenever the battery level on the peripheral is updated, whether
        it is by notification or from reading the characteristic itself.

        The DecodedReadCompleteEventArgs value given is the integer battery percent received. If the read failed
        or failed to decode, the value will be equal to the raw bytes received.
        """
        return self._on_battery_level_updated_event

    @property
    def can_enable_notifications(self) -> bool:
        """
        Checks if the battery level characteristic allows notifications to be subscribed to

        :return: True if notifications can be enabled, False if not
        """
        return self._batt_characteristic.subscribable

    def enable_notifications(self):
        """
        Enables notifications for the battery level characteristic. Note: this function will raise an exception
        if notifications aren't possible

        :return: a Waitable which waits for the write to finish
        """
        return self._batt_characteristic.subscribe(self._on_battery_level_notification)

    def disable_notifications(self):
        """
        Disables notifications for the battery level characteristic. Note: this function will raise an exception
        if notifications aren't possible

        :return: a Waitable which waits for the write to finish
        """
        return self._batt_characteristic.unsubscribe()

    def _on_battery_level_notification(self, characteristic, event_args):
        """
        :param characteristic:
        :type event_args: blatann.event_args.NotificationReceivedEventArgs
        :return:
        """
        decoded_value = None
        try:
            stream = ble_data_types.BleDataStream(event_args.value)
            decoded_value = BatteryLevel.decode(stream)
        except Exception as e:  # TODO not so generic
            logger.error("Failed to decode Battery Level, stream: [{}]".format(binascii.hexlify(event_args.value)))
            logger.exception(e)

        decoded_event_args = DecodedReadCompleteEventArgs.from_notification_complete_event_args(event_args, decoded_value)
        self._on_battery_level_updated_event.notify(self, decoded_event_args)

    def _on_read_complete(self, characteristic, event_args):
        """
        :param characteristic:
        :type event_args: blatann.event_args.ReadCompleteEventArgs
        """
        decoded_value = None
        if event_args.status == GattStatusCode.success:
            try:
                stream = ble_data_types.BleDataStream(event_args.value)
                decoded_value = BatteryLevel.decode(stream)
            except Exception as e:  # TODO not so generic
                logger.error("Failed to decode Battery Level, stream: [{}]".format(binascii.hexlify(event_args.value)))
                logger.exception(e)

        decoded_event_args = DecodedReadCompleteEventArgs.from_read_complete_event_args(event_args, decoded_value)
        self._on_battery_level_updated_event.notify(self, decoded_event_args)

    @classmethod
    def find_in_database(cls, gattc_database):
        """
        :type gattc_database: blatann.gatt.gattc.GattcDatabase
        :rtype: BatteryClient
        """
        service = gattc_database.find_service(BATTERY_SERVICE_UUID)
        if service:
            return BatteryClient(service)



from __future__ import absolute_import
import glob
from os import path
import sys
import blatann.examples

# Non-example python files to exclude
EXCLUDES = ["__init__", "__main__", "constants", "example_utils"]

_py_files = [path.splitext(path.basename(f))[0] for f in glob.glob(path.dirname(__file__) + "/*.py")]
examples = [f for f in _py_files if f not in EXCLUDES]


def print_help():
    print("\nUsage: python -m blatann.examples [example_filename] [comport]")
    print("Examples:")
    for e in examples:
        print("  {}".format(e))
    sys.exit(1)


def main():
    if len(sys.argv) < 3:
        print_help()

    example_name = sys.argv[1]
    comport = sys.argv[2]

    if example_name not in examples:
        print("Unknown example {}".format(example_name))
        print_help()

    module = __import__(blatann.examples.__name__, fromlist=[example_name])
    example = getattr(module, example_name)
    example.main(comport)


if __name__ == '__main__':
    main()



import binascii
from blatann.gatt import GattStatusCode
from blatann.services import ble_data_types
from blatann.event_args import NotificationReceivedEventArgs, ReadCompleteEventArgs, DecodedReadCompleteEventArgs, \
    WriteEventArgs, DecodedWriteEventArgs


class DecodedReadWriteEventDispatcher(object):
    def __init__(self, owner, ble_type, event_to_raise, logger=None):
        self.owner = owner
        self.ble_type = ble_type
        self.event_to_raise = event_to_raise
        self.logger = logger

    def decode(self, data):
        try:
            stream = ble_data_types.BleDataStream(data)
            return self.ble_type.decode(stream)
        except Exception as e:
            if self.logger:
                self.logger.error("Failed to decode into {} type, stream: [{}]".format(self.ble_type.__name__,
                                                                                       binascii.hexlify(data)))
                self.logger.exception(e)
        return None

    def __call__(self, characteristic, event_args):
        if isinstance(event_args, ReadCompleteEventArgs):
            if event_args.status == GattStatusCode.success:
                decoded_value = self.decode(event_args.value)
            else:
                decoded_value = None
            decoded_event_args = DecodedReadCompleteEventArgs.from_read_complete_event_args(event_args, decoded_value)

        elif isinstance(event_args, NotificationReceivedEventArgs):
            decoded_value = self.decode(event_args.value)
            decoded_event_args = DecodedReadCompleteEventArgs.from_notification_complete_event_args(event_args,
                                                                                                    decoded_value)
        elif isinstance(event_args, WriteEventArgs):
            decoded_value = self.decode(event_args.value)
            decoded_event_args = DecodedWriteEventArgs(decoded_value, event_args.value)
        else:
            if self.logger:
                self.logger.error("Unable to handle unknown event args {}".format(event_args))
            return

        self.event_to_raise.notify(self.owner, decoded_event_args)




"""
This example demonstrates reading a peripheral's Device Info Service using blatann's device_info service module.
The operations here are programmed in a procedural manner.

This can be used alongside any peripheral which implements the DIS and advertises the 16-bit DIS service UUID.
The peripheral_device_info_service example can be used with this.
"""
from blatann import BleDevice
from blatann.examples import example_utils
from blatann.services import device_info
from blatann.nrf import nrf_events


logger = example_utils.setup_logger(level="INFO")


def main(serial_port):
    # Open the BLE Device and suppress spammy log messages
    ble_device = BleDevice(serial_port)
    ble_device.event_logger.suppress(nrf_events.GapEvtAdvReport)
    ble_device.open()

    # Set scan duration to 4 seconds
    ble_device.scanner.set_default_scan_params(timeout_seconds=4)
    logger.info("Scanning for peripherals advertising UUID {}".format(device_info.DIS_SERVICE_UUID))

    target_address = None
    # Start scan and wait for it to complete
    scan_report = ble_device.scanner.start_scan().wait()
    # Search each peer's advertising data for the DIS Service UUID to be advertised
    for report in scan_report.advertising_peers_found:
        if device_info.DIS_SERVICE_UUID in report.advertise_data.service_uuid16s:
            target_address = report.peer_address
            break

    if not target_address:
        logger.info("Did not find peripheral advertising DIS service")
        return

    # Initiate connection and wait for it to finish
    logger.info("Found match: connecting to address {}".format(target_address))
    peer = ble_device.connect(target_address).wait()
    if not peer:
        logger.warning("Timed out connecting to device")
        return

    logger.info("Connected, conn_handle: {}".format(peer.conn_handle))
    # Initiate service discovery and wait for it to complete
    _, event_args = peer.discover_services().wait(10, exception_on_timeout=False)
    logger.info("Service discovery complete! status: {}".format(event_args.status))

    # Find the device info service in the peer's database
    dis = device_info.find_device_info_service(peer.database)
    if not dis:
        logger.info("Failed to find Device Info Service in peripheral database")
        peer.disconnect().wait()
        return

    # Example 1:
    # Iterate through all possible device info characteristics, read the value if present in service
    for char in device_info.CHARACTERISTICS:
        if dis.has(char):
            logger.info("Reading characteristic: {}...".format(char))
            char, event_args = dis.get(char).wait()
            if isinstance(event_args.value, bytes):
                value = event_args.value.decode("utf8")
            else:
                value = event_args.value
            logger.info("Value: {}".format(value))

    # Example 2:
    # Read specific characteristics, if present in the service
    if dis.has_software_revision:
        char, event_args = dis.get_software_revision().wait()
        sw_version = event_args.value
        logger.info("Software Version: {}".format(sw_version.decode("utf8")))
    if dis.has_pnp_id:
        char, event_args = dis.get_pnp_id().wait()
        pnp_id = event_args.value  # type: device_info.PnpId
        logger.info("Vendor ID: {}".format(pnp_id.vendor_id))
    if dis.has_system_id:
        char, event_args = dis.get_system_id().wait()
        system_id = event_args.value  # type: device_info.SystemId
        logger.info("System ID: {}".format(system_id))

    # Disconnect and close device
    peer.disconnect().wait()
    ble_device.close()


if __name__ == '__main__':
    main("COM9")



from blatann.nrf.nrf_types import BLEHci as _BLEHci
from blatann.gap.smp import SecurityStatus, IoCapabilities, AuthenticationKeyType, SecurityParameters
from blatann.gap.scanning import ScanParameters
from blatann.gap.advertising import AdvertisingData, AdvertisingFlags

HciStatus = _BLEHci

"""
The default link-layer packet size used when a connection is established
"""
DLE_SIZE_DEFAULT = 27

"""
The minimum allowed link-layer packet size
"""
DLE_SIZE_MINIMUM = 27

"""
The maximum allowed link-layer packet size
"""
DLE_SIZE_MAXIMUM = 251



from blatann.uuid import Uuid16


GLUCOSE_SERVICE_UUID = Uuid16(0x1808)

MEASUREMENT_CHARACTERISTIC_UUID = Uuid16(0x2A18)
MEASUREMENT_CONTEXT_CHARACTERISTIC_UUID = Uuid16(0x2A34)
FEATURE_CHARACTERISTIC_UUID = Uuid16(0x2A51)
RACP_CHARACTERISTIC_UUID = Uuid16(0x2A52)

